/*
 *  Driver for AMD InterWave soundcard
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define SND_MAIN_OBJECT_FILE
#include "driver.h"
#include "pcm.h"
#include "mixer.h"
#include "gus.h"
#include "cs4231.h"

char *snd_id = NULL;		/* ID for this card */
int snd_port = SND_AUTO;	/* 0x210,0x220,0x230,0x240,0x250,0x260 */
int snd_irq = SND_AUTO;		/* 0,1,3,5,9,11,12,15 */
int snd_dma1 = SND_AUTO;	/* 0,1,3,5,6,7 */
int snd_dma2 = SND_AUTO;	/* 0,1,3,5,6,7 (-1 = disable) */
int snd_dma1_size = SND_AUTO; 	/* 8,16,32,64,128 */
int snd_dma2_size = SND_AUTO; 	/* 8,16,32,64,128 */
int snd_joystick_dac = 29;	/* 0 to 31, (0.59V-4.52V or 0.389V-2.98V) */
#ifdef MODULE_PARM
MODULE_PARM( snd_id, "s" );
MODULE_PARM_DESC( snd_id, "ID string for InterWave soundcard." );
MODULE_PARM( snd_port, "i" );
MODULE_PARM_DESC( snd_port, "Port # for InterWave driver." );
MODULE_PARM( snd_irq, "i" );
MODULE_PARM_DESC( snd_irq, "IRQ # for InterWave driver." );
MODULE_PARM( snd_dma1, "i" );
MODULE_PARM_DESC( snd_dma1, "DMA1 # for InterWave driver." );
MODULE_PARM( snd_dma2, "i" );
MODULE_PARM_DESC( snd_dma2, "DMA2 # for InterWave driver." );
MODULE_PARM( snd_dma1_size, "i" );
MODULE_PARM_DESC( snd_dma1_size, "DMA1 size in kB for InterWave driver." );
MODULE_PARM( snd_dma2_size, "i" );
MODULE_PARM_DESC( snd_dma2_size, "DMA2 size in kB for InterWave driver." );
MODULE_PARM( snd_joystick_dac, "i" );
MODULE_PARM_DESC( snd_joystick_dac, "Joystick DAC level 0.59V-4.52V or 0.389V-2.98V for InterWave driver." );
#endif

static int snd_irqnum = -1;
static int snd_dma1num = -1;
static int snd_dma2num = -1;

static snd_card_t *snd_card = NULL;
static snd_gus_card_t *snd_gus = NULL;
static snd_pcm_t *snd_pcm = NULL;
static snd_mixer_t *snd_mixer = NULL;

static unsigned short snd_gus_status;
static unsigned short snd_pcm_status;

static void snd_interwave_use_inc( void )
{
  MOD_INC_USE_COUNT;
}

static void snd_interwave_use_dec( void )
{
  MOD_DEC_USE_COUNT;
}

static int snd_interwave_detect( snd_gus_card_t *gus, unsigned short port )
{
  unsigned long flags;
  unsigned char rev1, rev2;

  if ( snd_register_ioport( gus -> card, port, 16, "InterWave (Adlib/SB compatibility)" ) < 0 )
    return -EBUSY;
  if ( snd_register_ioport( gus -> card, port + 0x100, 12, "InterWave (synthesizer)" ) < 0 ) {
    snd_unregister_ioports( gus -> card );
    return -EBUSY;
  }
  if ( snd_register_ioport( gus -> card, port + 0x10c, 4, "InterWave (CS4231)" ) < 0 ) {
    snd_unregister_ioports( gus -> card );
    return -EBUSY;
  }
  snd_gus_set_port( gus, port );
  snd_gf1_i_write8( gus, SND_GF1_GB_RESET, 0 );	/* reset GF1 */
#ifdef ULTRACFG_DEBUG_DETECT
  if ( ((d = snd_gf1_i_look8( gus, SND_GF1_GB_RESET )) & 0x07) != 0 ) {
    PRINTK( "ultra: [0x%x] check 1 failed - 0x%x\n", gus -> gf1.port, d );
    goto __nodev;
  }
#else
  if ( (snd_gf1_i_look8( gus, SND_GF1_GB_RESET ) & 0x07) != 0 ) goto __nodev;
#endif
  snd_delay( 16 );
  snd_gf1_i_write8( gus, SND_GF1_GB_RESET, 1 );	/* release reset */
  snd_delay( 16 );
#ifdef ULTRACFG_DEBUG_DETECT
  if ( ((d = snd_gf1_i_look8( gus, SND_GF1_GB_RESET )) & 0x07) != 1 ) {
    PRINTK( "ultra: [0x%x] check 2 failed - 0x%x\n", gus -> gf1.port, d );
    goto __nodev;
  }
#else
  if ( (snd_gf1_i_look8( gus, SND_GF1_GB_RESET ) & 0x07) != 1 ) goto __nodev;
#endif

  CLI( &flags );
  rev1 = snd_gf1_look8( gus, SND_GF1_GB_VERSION_NUMBER );
  snd_gf1_write8( gus, SND_GF1_GB_VERSION_NUMBER, ~rev1 );
  rev2 = snd_gf1_look8( gus, SND_GF1_GB_VERSION_NUMBER );
  snd_gf1_write8( gus, SND_GF1_GB_VERSION_NUMBER, rev1 );
  STI( &flags );
#ifdef ULTRACFG_DEBUG_DETECT
  PRINTK( "ultra: [0x%x] InterWave check - rev1=0x%x, rev2=0x%x\n", gus -> gf1.port, rev1, rev2 );
#endif
  if ( ( rev1 & 0xf0 ) == ( rev2 & 0xf0 ) &&
       ( rev1 & 0x0f ) != ( rev2 & 0x0f ) )
    {
#ifdef ULTRACFG_DEBUG_DETECT
      PRINTK( "ultra: [0x%x] InterWave check - passed\n", gus -> gf1.port );
#endif
      gus -> interwave = 1;
      return 0;			/* ok.. We have InterWave board */
    }
#ifdef ULTRACFG_DEBUG_DETECT
  PRINTK( "ultra: [0x%x] InterWave check - failed\n", gus -> gf1.port );
#endif
  __nodev:
  snd_unregister_ioports( gus -> card );
  return -ENODEV;
}

static void snd_interwave_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
  register int loop;
  register unsigned short status;

  if ( !snd_gus || !snd_pcm ) return;

  do {
    loop = 0;
    if ( (status = INB( snd_gus_status )) ) {
      snd_gus_interrupt( snd_gus, status );
      loop++;
    }
    status = INB( snd_pcm_status );
    if ( (status = INB( snd_pcm_status )) & 0x01 ) {	/* IRQ bit is set? */
      snd_cs4231_interrupt( snd_pcm, status );
      loop++;
    }
  } while ( loop );
}

static int snd_interwave_resources( snd_card_t *card )
{
  static int possible_irqs[] = { 5, 11, 12, 9, 7, 15, 3, 4, -1 };
  static int possible_dmas[] = { 0, 1, 3, 5, 6, 7, -1 };

  if ( (snd_irqnum = snd_register_interrupt( card, "InterWave", snd_irq, snd_interwave_interrupt, possible_irqs )) < 0 ) {
    return snd_irqnum;
  }
  if ( (snd_dma1num = snd_register_dma_channel( card, "GFA1/CS4231 record", snd_dma1, snd_dma1_size, possible_dmas )) < 0 ) {
    return snd_dma1num;
  }
  if ( snd_dma2 >= 0 ) {
    if ( (snd_dma2num = snd_register_dma_channel( card, "CS4231 playback", snd_dma2, snd_dma2_size, possible_dmas )) < 0 ) {
      return snd_dma2num;
    }
  }
  return 0;
}

static void snd_interwave_reset( snd_gus_card_t *gus )
{
  snd_gf1_write8( gus, SND_GF1_GB_RESET, 0x00 );
  snd_delay( 16 );
  snd_gf1_write8( gus, SND_GF1_GB_RESET, 0x01 );
  snd_delay( 16 );
}

static void snd_interwave_bank_sizes( snd_gus_card_t *gus, int *sizes )
{
  int idx;
  unsigned int local;
  unsigned char d;

  for ( idx = 0; idx < 4; idx++ )
    {
      sizes[ idx ] = 0;
      d = 0x55;
      for ( local = idx << 22;
            local < ( idx << 22 ) + 0x400000;
            local += 0x40000, d++ )
        {
          snd_gf1_poke( gus, local, d );
          snd_gf1_poke( gus, local + 1, d + 1 );
#if 0
          printk( "d = 0x%x, local = 0x%x, local + 1 = 0x%x, idx << 22 = 0x%x\n",
          		d,
                        snd_gf1_peek( gus, local ),
                        snd_gf1_peek( gus, local + 1 ),
                        snd_gf1_peek( gus, idx << 22 ) );
#endif
          if ( snd_gf1_peek( gus, local ) != d ||
               snd_gf1_peek( gus, local + 1 ) != d + 1 ||
               snd_gf1_peek( gus, idx << 22 ) != 0x55 ) break;
          sizes[ idx ]++;
        }
    }
#if 0
  printk( "sizes: %i %i %i %i\n", sizes[ 0 ], sizes[ 1 ], sizes[ 2 ], sizes[ 3 ] );
#endif
}

struct rom_hdr {
  /* 000 */ unsigned char iwave[ 8 ];
  /* 008 */ unsigned char rom_hdr_revision;
  /* 009 */ unsigned char series_number;
  /* 010 */ unsigned char series_name[ 16 ];
  /* 026 */ unsigned char date[ 10 ];
  /* 036 */ unsigned short vendor_revision_major;
  /* 038 */ unsigned short vendor_revision_minor;
  /* 040 */ unsigned int rom_size;
  /* 044 */ unsigned char copyright[ 128 ];
  /* 172 */ unsigned char vendor_name[ 64 ];
  /* 236 */ unsigned char rom_description[ 128 ];
  /* 364 */ unsigned char pad[ 147 ];
  /* 511 */ unsigned char csum;
};

static void snd_interwave_detect_memory( snd_gus_card_t *gus )
{
  static unsigned int lmc[ 13 ] = {
    0x00000001, 0x00000101, 0x01010101, 0x00000401,
    0x04040401, 0x00040101, 0x04040101, 0x00000004,
    0x00000404, 0x04040404, 0x00000010, 0x00001010,
    0x10101010
  };

  int i, bank_pos, pages;
  unsigned int lmct;
  int psizes[ 4 ];
  unsigned char csum;
  struct rom_hdr romh;

  gus -> card -> type = SND_CARD_TYPE_AMD_INTERWAVE;
  snd_interwave_reset( gus );
  snd_gf1_write8( gus, SND_GF1_GB_GLOBAL_MODE, snd_gf1_read8( gus, SND_GF1_GB_GLOBAL_MODE ) | 0x01 );	/* enhanced mode */
  snd_gf1_write8( gus, SND_GF1_GB_MEMORY_CONTROL, 0x01 );  /* DRAM I/O cycles selected */
  snd_gf1_write16( gus, SND_GF1_GW_MEMORY_CONFIG, ( snd_gf1_look16( gus, SND_GF1_GW_MEMORY_CONFIG ) & 0xff10 ) | 0x004c );
  /* ok.. simple test of memory size */
  pages = 0;
  snd_gf1_poke( gus, 0, 0x55 );
  snd_gf1_poke( gus, 1, 0xaa );
#if 1
  if ( snd_gf1_peek( gus, 0 ) == 0x55 && snd_gf1_peek( gus, 1 ) == 0xaa )
#else
  if ( 0 )		/* ok.. for testing of 0k RAM */
#endif
    {
      snd_interwave_bank_sizes( gus, psizes );
      lmct = ( psizes[ 3 ] << 24 ) | ( psizes[ 2 ] << 16 ) |
             ( psizes[ 1 ] << 8 ) | psizes[ 0 ];
#if 0
      printk( "lmct = 0x%08x\n", lmct );
#endif
      for ( i = 0; i < sizeof( lmc ) / sizeof( unsigned int ); i++ )
        if ( lmct == lmc[ i ] )
          {
#if 0
            printk( "found !!! %i\n", i );
#endif
            snd_gf1_write16( gus, SND_GF1_GW_MEMORY_CONFIG, ( snd_gf1_look16( gus, SND_GF1_GW_MEMORY_CONFIG ) & 0xfff0 ) | i );
            snd_interwave_bank_sizes( gus, psizes );
            break;
          }
      if ( i >= sizeof( lmc ) / sizeof( unsigned int ) && !gus -> gf1.enh_mode )
        snd_gf1_write16( gus, SND_GF1_GW_MEMORY_CONFIG, ( snd_gf1_look16( gus, SND_GF1_GW_MEMORY_CONFIG ) & 0xfff0 ) | 2 );
      for ( i = 0; i < 4; i++ )
        {
          gus -> gf1.banks_8[ i ].address =
          gus -> gf1.banks_16[ i ].address = i << 22;
          gus -> gf1.banks_8[ i ].size =
          gus -> gf1.banks_16[ i ].size = psizes[ i ] << 18;
          pages += psizes[ i ];
        }
    }
  pages <<= 18;
  gus -> gf1.memory = pages;
  
  snd_gf1_write8( gus, SND_GF1_GB_MEMORY_CONTROL, 0x03 );  /* select ROM */
  snd_gf1_write16( gus, SND_GF1_GW_MEMORY_CONFIG, ( snd_gf1_look16( gus, SND_GF1_GW_MEMORY_CONFIG ) & 0xff1f ) | ( 4 << 5 ) );
  gus -> gf1.rom_banks = 0;
  gus -> gf1.rom_memory = 0;
  for ( bank_pos = 0; bank_pos < 16L * 1024L * 1024L; bank_pos += 4L * 1024L * 1024L )
    {
      for ( i = 0; i < sizeof( struct rom_hdr ); i++ )
        *(((unsigned char *)&romh) + i) = snd_gf1_peek( gus, i + bank_pos );
#ifdef ULTRACFG_DEBUG_ROM
      printk( "ultra: ROM at 0x%06x = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x\n", bank_pos,
        		romh.iwave[ 0 ], romh.iwave[ 1 ], romh.iwave[ 2 ], romh.iwave[ 3 ],
        		romh.iwave[ 4 ], romh.iwave[ 5 ], romh.iwave[ 6 ], romh.iwave[ 7 ] );
#endif
      if ( strncmp( romh.iwave, "INTRWAVE", 8 ) ) continue;  /* first check */
      csum = 0;
      for ( i = 0; i < sizeof( struct rom_hdr ) - 1; i++ )
        csum += *(((unsigned char *)&romh) + i); 
#ifdef ULTRACFG_DEBUG_ROM
      printk( "ultra: ROM checksum = 0x%x == 0x%x (computed)\n", romh.csum, (unsigned char)(256 - csum) );
#endif
      if ( 256 - csum != romh.csum ) continue;		  /* not valid rom */
      gus -> gf1.rom_banks++;
      gus -> gf1.rom_present |= 1 << ( bank_pos >> 22 );
      gus -> gf1.rom_memory = snd_get_dword( (char *)&romh.rom_size, 0 );
    }
#if 0
  if ( gus -> gf1.rom_memory > 0 )
    {
      if ( gus -> gf1.rom_banks == 1 && gus -> gf1.rom_present == 8 )
        gus -> card -> version = SND_CARD_TYPE_IW_DYNASONIC;
    }
#endif
  snd_gf1_write8( gus, SND_GF1_GB_MEMORY_CONTROL, 0x00 );  /* select RAM */
    
  if ( !gus -> gf1.enh_mode ) snd_interwave_reset( gus );
}

static void snd_interwave_init( snd_gus_card_t *gus )
{
  unsigned long flags;

  /* ok.. some InterWave specific initialization */
  CLI( &flags );
  snd_gf1_write8( gus, SND_GF1_GB_SOUND_BLASTER_CONTROL, 0x00 );
  snd_gf1_write8( gus, SND_GF1_GB_COMPATIBILITY, 0x1f );
  snd_gf1_write8( gus, SND_GF1_GB_DECODE_CONTROL, 0x49 );
  snd_gf1_write8( gus, SND_GF1_GB_VERSION_NUMBER, 0x11 );
  snd_gf1_write8( gus, SND_GF1_GB_MPU401_CONTROL_A, 0x00 );
  snd_gf1_write8( gus, SND_GF1_GB_MPU401_CONTROL_B, 0x30 );
  snd_gf1_write8( gus, SND_GF1_GB_EMULATION_IRQ, 0x00 );
  STI( &flags );
  gus -> equal_irq = 1;
  gus -> codec_flag = 1;
  gus -> interwave = 1;
  gus -> max_flag = 1;
  gus -> joystick_dac = snd_joystick_dac;
}

#define CS4231_PRIVATE( left, right, shift, mute ) ((left << 24)|(right << 16)|(shift<<8)|mute)

static int snd_interwave_mixer( snd_mixer_t *mixer )
{
  static struct SND_STRU_MIXER_CHANNEL_HW master = {
    SND_MIXER_PRI_MASTER,		/* priority */
    SND_MIXER_ID_MASTER,		/* device name */
    SND_MIXER_VOLUME,			/* OSS device # */
    1, 1, 1,				/* mute/stereo/record */
    31,					/* max. value */
    CS4231_PRIVATE( CS4231_LINE_LEFT_OUTPUT, CS4231_LINE_RIGHT_OUTPUT, 0, 0x80 ),
    NULL,				/* compute volume */
    NULL,				/* record source */
    NULL,				/* set mute */
    NULL,				/* set volume level */
  };
  snd_mixer_channel_t *channel;
  int idx;
  
  /* ok. InterWave have MIC different (stereo) */
  channel = NULL;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    switch ( channel -> hw.priority ) {
      case SND_MIXER_PRI_MIC:
        channel -> hw.stereo = 1;
        channel -> hw.max = 31;
        channel -> hw.private_value = CS4231_PRIVATE( CS4231_LEFT_MIC_INPUT, CS4231_RIGHT_MIC_INPUT, 0, 0x80 );
        break;
      case SND_MIXER_PRI_AUXA:
        channel -> hw.priority = SND_MIXER_PRI_SYNTHESIZER;
        channel -> hw.ossdev = SND_MIXER_SYNTH;
        strcpy( channel -> hw.name, SND_MIXER_ID_SYNTHESIZER );
        snd_mixer_reorder_channel( mixer, channel );
        break;
      case SND_MIXER_PRI_AUXB:
        channel -> hw.priority = SND_MIXER_PRI_CD;
        channel -> hw.ossdev = SND_MIXER_CD;
        strcpy( channel -> hw.name, SND_MIXER_ID_CD );
        snd_mixer_reorder_channel( mixer, channel );
        break;
    }
  }
  if ( !channel ) return -EINVAL;
  /* add master volume control */
  master.set_record_source = channel -> hw.set_record_source;
  master.set_mute = channel -> hw.set_mute;
  master.set_volume_level = channel -> hw.set_volume_level;
  channel = snd_mixer_new_channel( mixer, &master );
  if ( !channel ) return -ENOMEM;
  return 0;
}

int init_module( void )
{
  static int possible_ports[] = { 0x210,0x220,0x230,0x240,0x250,0x260, -1 };
  int *ports = possible_ports;
    
  snd_card = snd_card_new( snd_id, "AMD InterWave", snd_interwave_use_inc, snd_interwave_use_dec );
  if ( !snd_card ) return -ENOMEM;
  if ( snd_interwave_resources( snd_card ) < 0 ) {
    snd_card_free( snd_card );
    return -EBUSY;
  }
  snd_gus = snd_gus_new_card( snd_card,
                              snd_port,
                              snd_irqnum,
                              snd_dma1num,
                              snd_dma2num );
  if ( !snd_gus ) {
    snd_card_free( snd_card );
    return -ENOMEM;
  }
  if ( snd_port == SND_AUTO ) {
    for ( ports = possible_ports; *ports >= 0; ports++ ) {
      if ( !snd_interwave_detect( snd_gus, *ports ) ) break;
    }
    if ( *ports < 0 ) {
      snd_card_free( snd_card );
      PRINTK( "ultra: InterWave soundcard not found or device busy\n" );
      return -ENODEV;
    }
  } else {
    if ( snd_interwave_detect( snd_gus, snd_port ) ) {
      snd_card_free( snd_card );
      PRINTK( "ultra: InterWave soundcard not found at 0x%x or device busy\n", snd_port );
      return -ENODEV;
    }
  }
  if ( snd_card_register( snd_card ) ) {
    snd_card_free( snd_card );
    return -ENOMEM;
  }
  snd_gus_status = snd_gus -> gf1.reg_irqstat;
  snd_pcm_status = snd_gus -> gf1.port + 0x10c + 2;
  snd_interwave_init( snd_gus );
  snd_interwave_detect_memory( snd_gus );
  if ( snd_gus_init_dma_irq( snd_gus, 1 ) < 0 ) {
    snd_card_free( snd_card );
    return -EINVAL;
  }
  snd_pcm = snd_cs4231_new_device( snd_card,
                                   snd_gus -> gf1.port + 0x10c,
                                   snd_irqnum,
                                   snd_dma1num,
                                   snd_dma2num < 0 ? snd_dma1num : snd_dma2num,
                                   CS4231_HW_INTERWAVE );
  if ( !snd_pcm ) {
    __return1:
    snd_gus_init_dma_irq( snd_gus, 0 );
    snd_card_free( snd_card );
    return -ENXIO;
  }
  if ( snd_cs4231_probe( snd_pcm ) < 0 ) {
    PRINTD( "ultra: Oops, seems that InterWave card doesn't have CODEC CS4231?\n" );
    goto __return1;
  }
  if ( snd_pcm_register( snd_pcm, 0 ) < 0 ) goto __return1;
  snd_mixer = snd_cs4231_new_mixer( snd_pcm );
  if ( !snd_mixer ) {
    __return2:
    snd_pcm_free( snd_pcm );
    goto __return1;
  }
  if ( snd_interwave_mixer( snd_mixer ) < 0 ) {
    __return3:
    snd_mixer_free( snd_mixer );
    goto __return2;
  }    
  if ( snd_mixer_register( snd_mixer, 0 ) < 0 ) {
    goto __return3;
  }
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_MASTER, 70, 70, 0 );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_SYNTHESIZER, 75, 75, 0 );
  snd_mixer_set_kernel_mute( snd_mixer, SND_MIXER_PRI_PCM, SND_MIX_MUTE );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_PCM, 85, 85, 0 );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_LINE, 70, 70, SND_MIXER_FLG_MUTE );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_MIC, 70, 70, SND_MIXER_FLG_MUTE );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_CD, 70, 70, SND_MIXER_FLG_MUTE );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_GAIN, 0, 0, 0 );
  snd_mixer_channel_init( snd_mixer, SND_MIXER_PRI_LOOPBACK, 0, 0, SND_MIXER_FLG_MUTE );
  snd_enable_irq( snd_card, snd_irqnum );
#if 0
  snd_cs4231_debug( (cs4231_t *)snd_pcm -> playback.hw.private_data );
#endif
  return 0;
}

void cleanup_module( void )
{
  if ( snd_gus ) {
    snd_gus_init_dma_irq( snd_gus, 0 );
    if ( snd_mixer )
      snd_mixer_unregister( snd_mixer );
    if ( snd_pcm )
      snd_pcm_unregister( snd_pcm );
    if ( snd_card )
      snd_card_unregister( snd_card );
  }
}
